/// @file MLP.hpp
/// @author Franck Dary
/// @version 2.0
/// @date 2018-08-03

#ifndef MLP__H
#define MLP__H

#include "NeuralNetwork.hpp"
#include "MLPBase.hpp"
#include "ProgramParameters.hpp"

/// @brief Multi Layer Perceptron.
///
/// It is capable of training itself given a batch of examples.\n
/// Once trained, it can also be used to predict the class of a certain input.
class MLP : public NeuralNetwork
{
  private :

  /// @brief The mlp that will be trained.
  MLPBase mlp;
  /// @brief The training algorithm that will be used.
  std::unique_ptr<dynet::Trainer> trainer;

  public :

  /// @brief initialize a new untrained MLP from a desired topology.
  ///
  /// topology example for 2 hidden layers : (150,RELU,0.3)(50,ELU,0.2)\n
  /// Of sizes 150 and 50, activation functions RELU and ELU, and dropout rates
  /// of 0.3 and 0.2.
  /// @param nbInputs The size of the input layer of the MLP.
  /// @param topology Description of each hidden Layer of the MLP.
  /// @param nbOutputs The size of the output layer of the MLP.
  void init(int nbInputs, const std::string & topology, int nbOutputs) override;
  /// @brief Construct a new MLP for training.
  MLP();
  /// @brief Read and construct a trained MLP from a file.
  ///
  /// The file must have been written by save.
  /// @param filename The file to read the MLP from.
  MLP(const std::string & filename);
  /// @brief Give a score to each possible class, given an input.
  ///
  /// @param fd The input to use.
  ///
  /// @return A vector containing one score per possible class.
  std::vector<float> predict(FeatureModel::FeatureDescription & fd) override;
  /// @brief Update the parameters according to the given gold class.
  ///
  /// @param fd The input to use.
  /// @param gold The gold class of this input.
  ///
  /// @return The loss.
  float update(FeatureModel::FeatureDescription & fd, int gold) override;
  /// @brief Get the loss according to the given gold class.
  ///
  /// @param fd The input to use.
  /// @param gold The gold class of this input.
  ///
  /// @return The loss.
  float update(FeatureModel::FeatureDescription & fd, const std::vector<float> & gold) override;
  /// @brief Get the loss according to the given gold class.
  ///
  /// @param fd The input to use.
  /// @param gold The gold class of this input.
  ///
  /// @return The loss.
  float getLoss(FeatureModel::FeatureDescription & fd, int gold) override;
  /// @brief Get the loss according to the given gold vector.
  ///
  /// @param fd The input to use.
  /// @param gold The gold vector for this input.
  ///
  /// @return The loss.
  float getLoss(FeatureModel::FeatureDescription & fd, const std::vector<float> & gold) override;
  /// @brief Save the MLP to a file.
  /// 
  /// @param filename The file to write the MLP to.
  void save(const std::string & filename) override;
  /// @brief Print the topology (Layers) of the MLP.
  ///
  /// @param output Where the topology will be printed.
  void printTopology(FILE * output) override;
  /// @brief Allocate the correct trainer type depending on the program parameters.
  ///
  /// @return A pointer to the newly allocated trainer.
  dynet::Trainer * createTrainer();
  void endOfIteration();
};

#endif
