#include "MLP.hpp"

MLP::MLP() : mlp("MLP")
{
  randomSeed = ProgramParameters::seed;
  trainer.reset(createTrainer());
  initDynet();
}

MLP::MLP(const std::string & filename) : mlp("MLP")
{
  randomSeed = ProgramParameters::seed;
  trainer.reset(createTrainer());
  initDynet();

  mlp.loadStruct(model, filename, 0);
  mlp.loadParameters(model, filename);
}

void MLP::init(int nbInputs, const std::string & topology, int nbOutputs)
{
  setBatchSize(0);
  mlp.init(model, nbInputs, topology, nbOutputs);
}

dynet::Trainer * MLP::createTrainer()
{
  auto optimizer = noAccentLower(ProgramParameters::optimizer);

  dynet::Trainer * trainer = nullptr;

  if (optimizer == "amsgrad")
    trainer = new dynet::AmsgradTrainer(model, ProgramParameters::learningRate, ProgramParameters::beta1, ProgramParameters::beta2, ProgramParameters::bias);
  else if (optimizer == "adam")
    trainer =  new dynet::AdamTrainer(model, ProgramParameters::learningRate, ProgramParameters::beta1, ProgramParameters::beta2, ProgramParameters::bias);
  else if (optimizer == "sgd")
    trainer =  new dynet::SimpleSGDTrainer(model, ProgramParameters::learningRate);
  else if (optimizer == "none")
    return nullptr;

  if (trainer)
  {
    trainer->sparse_updates_enabled = true;
    return trainer;
  }

  fprintf(stderr, "ERROR (%s) : unknown optimizer \'%s\'. Aborting.\n", ERRINFO, optimizer.c_str());

  exit(1);

  return nullptr;
}

std::vector<float> MLP::predict(FeatureModel::FeatureDescription & fd)
{
  return mlp.predict(fd);
}

float MLP::update(FeatureModel::FeatureDescription & fd, int gold)
{
  mlp.setBatchSize(getBatchSize());
  try
  {
    float loss = mlp.update(fd, gold);
    trainer->update();
    return loss;
  } catch (BatchNotFull &)
  {
    return 0.0;
  }
}

float MLP::update(FeatureModel::FeatureDescription & fd, const std::vector<float> & gold)
{
  mlp.setBatchSize(getBatchSize());
  try
  {
    float loss = mlp.update(fd, gold);
    trainer->update();
    return loss;
  } catch (BatchNotFull &)
  {
    return 0.0;
  }
}

float MLP::getLoss(FeatureModel::FeatureDescription & fd, int gold)
{
  mlp.setBatchSize(getBatchSize());
  try
  {
    float loss = mlp.getLoss(fd, gold);
    return loss;
  } catch (BatchNotFull &)
  {
    return 0.0;
  }
}

float MLP::getLoss(FeatureModel::FeatureDescription & fd, const std::vector<float> & gold)
{
  mlp.setBatchSize(getBatchSize());
  try
  {
    float loss = mlp.getLoss(fd, gold);
    return loss;
  } catch (BatchNotFull &)
  {
    return 0.0;
  }
}

void MLP::save(const std::string & filename)
{
  File * file = new File(filename, "w");
  delete file;
  mlp.saveStruct(filename);
  mlp.saveParameters(filename);
}

void MLP::printTopology(FILE * output)
{
  mlp.printTopology(output);
}

void MLP::endOfIteration()
{
  mlp.endOfIteration();
}

