/// @file ActionBank.hpp
/// @author Franck Dary
/// @version 1.0
/// @date 2018-08-06

#ifndef ACTIONBANK__H
#define ACTIONBANK__H

#include "Action.hpp"

/// @brief Retrieve a sequence of BasicAction from an action name.
///
/// The purpose of this class is to parse an action name and use it to construct 
/// the corresponding sequence of BasicAction.\n
/// The caller will then probably use this sequence of BasicAction to create an Action.
class ActionBank
{
  public :

  /// @brief Construct a sequence of BasicAction from an action name.
  ///
  /// The creation is done dynamically, and if the name given is not correct,
  /// the program will abort. 
  /// @param name The name of the action (e.g. 'WRITE 0 POS adj').
  ///
  /// @return The corresponding sequence of BasicAction.
  static std::vector<Action::BasicAction> str2sequence(const std::string & name);
  /// @brief Return the lenght of the link that the action will create. Only used in error analysis.
  ///
  /// For instance the link length for dependency parsing will be the absolute distance between the head and the dependent.
  ///
  /// @param c Current Config
  /// @param action The name of the action that will create the link
  ///
  /// @return The length of the link the action will create, default value 0 if not appliable.
  static int getLinkLength(const Config & c, const std::string & action);

  private :

  /// @brief Write a string into a specific cell of the multi-tapes buffer.
  ///
  /// This is a helper function that helps construct BasicAction.
  /// @param config The Config which buffer will be written into.
  /// @param tapeName The name of the tape of the buffer that will be written into.
  /// @param value The value that will be written.
  /// @param relativeIndex The index of the column that will be written into, relatively to the head of the Config.
  static void simpleBufferWrite(Config & config, const std::string & tapeName, 
    const std::string & value, int relativeIndex);

  /// @brief Whether or not a call to simpleBufferWrite is possible.
  ///
  /// This is a helper function that helps construct BasicAction.
  /// @param config The Config simpleBufferWrite would be called with.
  /// @param tapeName The name of the tape simpleBufferWrite would be called with.
  /// @param relativeIndex The relative index simpleBufferWrite would be called with.
  ///
  /// @return Whether or not a call to simpleBufferWrite is possible.
  static bool simpleBufferWriteAppliable(Config & config,
    const std::string & tapeName, int relativeIndex);

  /// @brief Whether or not the transform rule is appliable.
  ///
  /// This is a helper function that helps construct BasicAction.
  /// @param config The current Config.
  /// @param tapeName The name of the tape the rule would be applied on.
  /// @param relativeIndex The relative index of the cell of the tape.
  /// @param rule The rule.
  ///
  /// @return Whether or not rule can be applied to the cell of tapeName.
  static bool isRuleAppliable(Config & config,
    const std::string & tapeName, int relativeIndex, const std::string & rule);

  /// @brief Apply a transformation rule to a copy of a multi-tapes buffer cell, and write the result in another cell.
  ///
  /// A rule is in the form '\@er\@w' which in this case means 'remove er suffix and add w suffix'
  /// @param config The Config which buffer will be written into.
  /// @param fromTapeName The name of the buffer's tape containing the string to copy and transform.
  /// @param targetTapeName The name of the buffer's tape the result will be written to.
  /// @param rule The rule to apply.
  /// @param relativeIndex The index of the column that will be read and written into, relatively to the head of the Config.
  static void writeRuleResult(Config & config, const std::string & fromTapeName, const std::string & targetTapeName, const std::string & rule, int relativeIndex);

  /// \brief Write something on the buffer
  ///
  /// \param tapeName The tape we will write to
  /// \param value The value we will write
  /// \param relativeIndex The write index relmative to the buffer's head
  ///
  /// \return A BasicAction doing all of that
  static Action::BasicAction bufferWrite(std::string tapeName, std::string value, int relativeIndex);

  /// \brief Move the buffer's head
  ///
  /// \param movement The relative movement of the buffer's head
  ///
  /// \return A BasicAction moving the head
  static Action::BasicAction moveHead(int movement);

  /// \brief Write something on the buffer
  ///
  /// \param tapeName The tape we will write to
  /// \param value The value we will write
  /// \param stackIndex The stack index of the buffer index
  ///
  /// \return A BasicAction doing all of that
  static Action::BasicAction stackWrite(std::string tapeName, std::string value, int stackIndex);

  /// \brief Push the head of the buffer into the stack
  ///
  /// \return The corresponding BasicAction
  static Action::BasicAction pushHead();

  /// \brief Pop the stack
  ///
  /// \return The corresponding BasicAction
  static Action::BasicAction stackPop(bool checkGov);
};

#endif
